from gp.entity_linking.candidate_recognition._heuristic_model import HeuristicCanReg
from gp.entity_linking.candidate_recognition._interface import ICanReg
from gp.entity_linking.candidate_recognition._oracle_model import OracleCanReg

__all__ = ["ICanReg", "HeuristicCanReg", "OracleCanReg"]
